# pylint: disable=missing-docstring
import os
import subprocess
import unittest
from unittest import mock

from django.conf import settings
from django.test import override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
from resolwe.flow.managers import manager
from resolwe.flow.models import Data, DataDependency, Process
from resolwe.test import (
    ProcessTestCase,
    TestCase,
    tag_process,
    with_docker_executor,
    with_null_executor,
)

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), "descriptors")


class GetToolsTestCase(TestCase):
    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.os")
    @mock.patch("resolwe.flow.utils.settings")
    def test_get_tools_paths(self, settings_mock, os_mock, apps_mock):
        apps_mock.get_app_configs.return_value = [
            mock.MagicMock(path="/resolwe/test_app1"),
            mock.MagicMock(path="/resolwe/test_app2"),
        ]
        os_mock.path.join = os.path.join
        os_mock.path.isdir.side_effect = [False, True]
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = ["/custom_tools"]

        base_executor = BaseFlowExecutorPreparer()
        tools_list = base_executor.get_tools_paths()

        self.assertEqual(len(tools_list), 2)
        self.assertIn("/resolwe/test_app2/tools", tools_list)
        self.assertIn("/custom_tools", tools_list)

    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.settings")
    def test_not_list(self, settings_mock, apps_mock):
        apps_mock.get_app_configs.return_value = []
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = "/custom_tools"

        base_executor = BaseFlowExecutorPreparer()
        with self.assertRaisesRegex(KeyError, "setting must be a list"):
            base_executor.get_tools_paths()


class ManagerRunProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR]
        )

    @tag_process("test-min")
    def test_minimal_process(self):
        self.run_process("test-min")

    @tag_process("test-missing-file")
    def test_missing_file(self):
        self.run_process("test-missing-file", assert_status=Data.STATUS_ERROR)

        data = Data.objects.last()
        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("Referenced file does not exist", data.process_error[0])

    @tag_process("test-spawn-new")
    def test_spawn(self):
        self.run_process("test-spawn-new", tags=["test-tag"])

        data = Data.objects.last()
        file_path = data.location.get_path(filename="foo.bar")
        self.assertEqual(data.output["saved_file"]["file"], "foo.bar")
        self.assertTrue(os.path.isfile(file_path))
        self.assertEqual(data.tags, ["test-tag"])

        parent_data = Data.objects.first()
        self.assertEqual(data.parents.count(), 1)
        self.assertEqual(data.parents.first(), parent_data)

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )

    @tag_process("test-spawn-missing-file")
    def test_spawn_missing_export(self):
        data = self.run_process(
            "test-spawn-missing-file", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(
            data.process_error[0], "Error while preparing spawned Data objects"
        )

    @tag_process(
        "test-broken",
        "test-broken-invalid-execution-engine",
        "test-broken-invalid-expression-engine",
        "test-broken-data-name",
    )
    def test_broken(self):
        Process.objects.create(
            slug="test-broken-invalid-execution-engine",
            name="Test Process",
            contributor=self.contributor,
            type="data:test",
            version=1,
            run={"language": "invalid",},
        )

        self.run_process("test-broken", assert_status=Data.STATUS_ERROR)
        self.run_process(
            "test-broken-invalid-expression-engine", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-broken-invalid-execution-engine", assert_status=Data.STATUS_ERROR
        )

        # If evaluation of data_name template fails, the process should not abort as the
        # template may be evaluatable later when the process completes.
        self.run_process("test-broken-data-name")

    @tag_process("test-broken-invalide-storage")
    def test_invalid_storage_file(self):
        data = self.run_process(
            "test-broken-invalide-storage", assert_status=Data.STATUS_ERROR
        )

        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertIn(
            "Value of 'storage' must be a valid JSON, current: 1a", data.process_error
        )

    @tag_process("test-workflow-1")
    def test_workflow(self):
        assign_perm("view_collection", self.contributor, self.collection)
        assign_perm("view_collection", self.group, self.collection)
        workflow_data = self.run_process(
            "test-workflow-1", {"param1": "world"}, tags=["test-tag"]
        )

        workflow_data.refresh_from_db()
        step1_data = Data.objects.get(process__slug="test-example-1")
        step2_data = Data.objects.get(process__slug="test-example-2")

        # Workflow should output indices of all data objects, in order.
        self.assertEqual(workflow_data.output["steps"], [step1_data.pk, step2_data.pk])

        # Steps should execute with the correct variables.
        self.assertEqual(step1_data.input["param1"], "world")
        self.assertEqual(step1_data.input["param2"], True)
        self.assertEqual(step1_data.output["out1"], "hello world")
        self.assertEqual(step1_data.tags, ["test-tag"])
        self.assertEqual(step2_data.input["param1"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["a"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["b"], "hello")
        self.assertEqual(step2_data.output["out1"], "simon says: hello world")
        self.assertEqual(step2_data.tags, ["test-tag"])

        self.assertEqual(step1_data.parents.count(), 1)
        self.assertEqual(step1_data.parents.first(), workflow_data)
        self.assertEqual(step2_data.parents.count(), 2)
        self.assertCountEqual(step2_data.parents.all(), [workflow_data, step1_data])

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in step1_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )
        self.assertEqual(
            {d.kind for d in step2_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS, DataDependency.KIND_IO},
        )

        self.assertTrue(self.contributor.has_perm("flow.view_data", step1_data))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm("flow.view_data", step1_data))

    @tag_process("test-workflow-2")
    def test_workflow_entity(self):
        with self.preparation_stage():
            # `self.run_process` adds input data in self.collection
            input_data = self.run_process("test-example-3")

        workflow = self.run_process("test-workflow-2", {"data1": input_data.pk})

        # Check that workflow results are added to the collection.
        self.assertEqual(
            self.collection.data.filter(pk__in=workflow.output["steps"]).count(), 1
        )
        # self.collection now contains workflow and two "normal" data objects
        self.assertEqual(self.collection.data.all().count(), 3)

    @with_docker_executor
    @tag_process("test-docker")
    def test_run_in_docker(self):
        data = self.run_process("test-docker")
        self.assertEqual(data.output["result"], "OK")

    @with_docker_executor
    @tag_process("test-requirements-docker")
    def test_executor_requirements(self):
        data = self.run_process("test-requirements-docker")
        self.assertEqual(data.output["result"], "OK")

    @with_docker_executor
    @tag_process("test-docker-uid-gid")
    def test_docker_uid_gid(self):
        data = self.run_process("test-docker-uid-gid")
        self.assertEqual(data.output["result"], "OK")

    @with_null_executor
    @tag_process("test-save-number")
    def test_null_executor(self):
        data = self.run_process(
            "test-save-number", {"number": 19}, assert_status=Data.STATUS_WAITING
        )
        self.assertEqual(data.input["number"], 19)
        self.assertEqual(data.output, {})

    # TODO: Debug why the 'test-memory-resource-alloc' process doesn't end with and error on Travis
    @unittest.skipIf(os.environ.get("TRAVIS", "") == "true", "Fails on Travis CI")
    @with_docker_executor
    @tag_process("test-memory-resource-alloc", "test-memory-resource-noalloc")
    def test_memory_resource(self):
        # This process should be terminated due to too much memory usage.
        self.run_process("test-memory-resource-alloc", assert_status=Data.STATUS_ERROR)
        # This process should run normally (honors the limits).
        self.run_process("test-memory-resource-noalloc")

    @with_docker_executor
    @tag_process("test-cpu-resource-1core", "test-cpu-resource-2core")
    def test_cpu_resource(self):
        # Currently there is no good way to test this reliably, so we just check if the
        # resource limit specification still makes the process run.
        self.run_process("test-cpu-resource-1core")
        self.run_process("test-cpu-resource-2core")

    @unittest.skipIf(
        settings.FLOW_DOCKER_DISABLE_SECCOMP, "Docker seccomp is disabled."
    )
    @with_docker_executor
    @tag_process(
        "test-network-resource-enabled",
        "test-network-resource-disabled",
        "test-network-resource-policy",
    )
    def test_network_resource(self):
        self.run_process("test-network-resource-enabled")
        self.run_process(
            "test-network-resource-disabled", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-network-resource-policy", assert_status=Data.STATUS_ERROR
        )

    @with_docker_executor
    @override_settings(FLOW_PROCESS_RESOURCE_DEFAULTS={"cpu_time_interactive": 1})
    @tag_process(
        "test-scheduling-class-interactive-ok",
        "test-scheduling-class-interactive-fail",
        "test-scheduling-class-batch",
    )
    def test_scheduling_class(self):
        self.run_process("test-scheduling-class-interactive-ok")
        self.run_process(
            "test-scheduling-class-interactive-fail", assert_status=Data.STATUS_ERROR
        )
        self.run_process("test-scheduling-class-batch")

    @with_docker_executor
    @tag_process("test-save-number")
    def test_executor_fs_lock(self):
        # First, run the process normaly.
        data = self.run_process("test-save-number", {"number": 42})

        # Make sure that process was successfully ran first time.
        self.assertEqual(data.output["number"], 42)
        data.output = {}
        data.save()

        process = subprocess.run(
            ["python", "-m", "executors", ".docker"],
            cwd=data.location.get_runtime_path(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=5,
        )

        # Drain control events generated by this executor.
        manager.reset(keep_state=True)

        self.assertEqual(process.returncode, 0)

        # Check the status of the data object.
        data.refresh_from_db()
        # Check that output is empty and thus process didn't ran.
        self.assertEqual(data.output, {})
        self.assertEqual(data.status, Data.STATUS_DONE)
        self.assertEqual(data.process_error, [])



# pylint: disable=missing-docstring
import contextlib
import os
import shutil
import tempfile
import unittest
from unittest.mock import MagicMock, patch

from testfixtures import LogCapture

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.utils.crypto import get_random_string

from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import Data, DataLocation, DescriptorSchema, Process, Storage
from resolwe.flow.utils import purge
from resolwe.test import ProcessTestCase


class PurgeTestFieldsMixin:
    """
    This class contains tests, which validate each field individually. It is used
    to perform basic unit tests (on simulated data, without running any processors)
    as well as more complex end-to-end tests (by running processors). Having them
    in this mixin makes sure that both, unit and e2e tests, are run for each field.
    """

    def test_basic_file(self):
        self.assertFieldWorks(
            "basic:file:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_specialization(self):
        self.assertFieldWorks(
            "basic:file:html:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_list(self):
        self.assertFieldWorks(
            "list:basic:file:",
            field_value=[{"file": "but_not_this"}, {"file": "and_not_this"}],
            script_setup="touch remove_this but_not_this and_not_this",
            script_save="re-save-file-list sample but_not_this and_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this", "and_not_this"],
        )

    def test_basic_dir(self):
        self.assertFieldWorks(
            "basic:dir:",
            field_value={"dir": "but_not_this"},
            script_setup="""
mkdir remove_this but_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
""",
            script_save="re-save-dir sample but_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=["but_not_this/a", "but_not_this/b", "but_not_this/c"],
        )

    def test_basic_dir_list(self):
        self.assertFieldWorks(
            "list:basic:dir:",
            field_value=[{"dir": "but_not_this"}, {"dir": "and_not_this"}],
            script_setup="""
mkdir remove_this but_not_this and_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
touch and_not_this/a and_not_this/b and_not_this/c
""",
            script_save="re-save-dir-list sample but_not_this and_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=[
                "but_not_this/a",
                "but_not_this/b",
                "but_not_this/c",
                "and_not_this/a",
                "and_not_this/b",
                "and_not_this/c",
            ],
        )


@override_settings(TEST_PROCESS_REQUIRE_TAGS=False)  # Test uses dynamic processes.
class PurgeE2ETest(PurgeTestFieldsMixin, ProcessTestCase):
    def create_and_run_processor(self, processor, **kwargs):
        processor_slug = get_random_string(6)
        Process.objects.create(
            slug=processor_slug,
            name="Test Purge Process",
            contributor=self.admin,
            type="data:test",
            version=1,
            **processor,
        )

        data = self.run_process(processor_slug, **kwargs)
        # Purge is normally called in an async worker, so we have to emulate the call.
        purge.location_purge(location_id=data.location.id, delete=True)

        return data

    def assertFilesRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertFalse(os.path.isfile(path), msg=path)

    def assertFilesNotRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertTrue(os.path.isfile(path), msg=path)

    def test_complex_processor(self):
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[
                    {
                        "name": "sample_file",
                        "label": "Sample output file",
                        "type": "basic:file:",
                    },
                    {
                        "name": "sample_dir",
                        "label": "Sample output directory",
                        "type": "basic:dir:",
                    },
                    {
                        "name": "sample_file_list",
                        "label": "Sample list of output files",
                        "type": "list:basic:file:",
                    },
                    {
                        "name": "sample_dir_list",
                        "label": "Sample list of output directories",
                        "type": "list:basic:dir:",
                    },
                ],
                run={
                    "language": "bash",
                    "program": """
touch these files should be removed
touch this_file_should_stay
mkdir -p directory/should/be
touch directory/should/be/removed
touch directory/should/be/removed2
mkdir -p stay/directory
touch stay/directory/file1
touch stay/directory/file2
touch entry1 entry2 entry3
mkdir dir1 dir2 dir3
touch dir1/a dir2/b dir3/c dir3/d
touch ref1 ref2 ref3 ref4
mkdir refs
touch refs/a refs/b

re-save-file sample_file this_file_should_stay ref3
re-save-dir sample_dir stay ref4
re-save-file-list sample_file_list entry1 entry2:ref1 entry3:refs
re-save-dir-list sample_dir_list dir1:ref2 dir2 dir3
""",
                },
            )
        )

        self.assertFilesRemoved(
            data, "these", "files", "should", "be", "removed", "directory"
        )
        self.assertFilesNotRemoved(
            data,
            "this_file_should_stay",
            "stay/directory/file1",
            "stay/directory/file2",
            "entry1",
            "entry2",
            "entry3",
            "dir1/a",
            "dir2/b",
            "dir3/c",
            "dir3/d",
            "ref1",
            "ref2",
            "ref3",
            "ref4",
            "refs/a",
            "refs/b",
        )

        data.location.refresh_from_db()
        self.assertEqual(data.location.purged, True)

    def assertFieldWorks(
        self, field_type, field_value, script_setup, script_save, removed, not_removed
    ):
        """
        Checks that a field is handled correctly when running a processor, which
        uses the field.
        """

        field_schema = {"name": "sample", "label": "Sample output", "type": field_type}

        # Test output.
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[field_schema],
                run={"language": "bash", "program": script_setup + "\n" + script_save},
            )
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

        # Test descriptor.
        descriptor_schema = DescriptorSchema.objects.create(
            slug=get_random_string(6), contributor=self.admin, schema=[field_schema]
        )
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[],
                run={"language": "bash", "program": script_setup},
            ),
            descriptor_schema=descriptor_schema,
            descriptor={"sample": field_value},
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

    @unittest.skipIf(
        True,
        "since PR308: manager now separated into parts, executor not logging here anymore",
    )
    def test_exception_logging(self):
        with patch("resolwe.flow.utils.purge.os", wraps=os) as os_mock:
            # Ensure that purge raises an exception, so we can check whether the exception
            # gets logged correctly.
            class TestPurgeException(Exception):
                pass

            def exception_raiser(*args, **kwargs):
                raise TestPurgeException

            os_mock.walk = exception_raiser

            with LogCapture() as log:
                self.create_and_run_processor(
                    processor=dict(
                        input_schema=[],
                        output_schema=[
                            {
                                "name": "sample",
                                "label": "Sample output",
                                "type": "basic:file:",
                            }
                        ],
                        run={
                            "language": "bash",
                            "program": "touch sample && re-save-file sample sample",
                        },
                    )
                )

                self.assertEqual(len(log.records), 1)
                self.assertEqual(log.records[0].name, "resolwe.flow.executors")
                self.assertTrue(str(log.records[0].msg).startswith("Purge error:"))
                self.assertTrue("TestPurgeException" in str(log.records[0].msg))


class PurgeUnitTest(PurgeTestFieldsMixin, ProcessTestCase):
    def assertFieldWorks(
        self, field_type, field_value, script_setup, script_save, removed, not_removed
    ):
        """
        Checks that a field is handled correctly by `get_purge_files` under a
        simulated Data object.
        """

        field_schema = {"name": "sample", "label": "Sample output", "type": field_type}

        # Test simulated operation.
        simulated_root = tempfile.mkdtemp()
        try:
            for filename in removed + not_removed:
                directory, basename = os.path.split(filename)
                if directory:
                    try:
                        os.makedirs(os.path.join(simulated_root, directory))
                    except OSError:
                        pass

                if basename:
                    with open(os.path.join(simulated_root, filename), "w"):
                        pass

            unreferenced_files = purge.get_purge_files(
                simulated_root,
                output={"sample": field_value},
                output_schema=[field_schema],
                descriptor={},
                descriptor_schema=[],
            )

            def strip_slash(filename):
                return filename[:-1] if filename[-1] == "/" else filename

            for filename in not_removed:
                self.assertNotIn(
                    strip_slash(os.path.join(simulated_root, filename)),
                    unreferenced_files,
                )

            for filename in removed:
                filename = strip_slash(os.path.join(simulated_root, filename))
                self.assertIn(filename, unreferenced_files)
                unreferenced_files.discard(filename)

            # Ensure that nothing more is removed.
            self.assertEqual(len(unreferenced_files), 0)
        finally:
            shutil.rmtree(simulated_root)

    def create_test_file(self, location, filename):
        """
        Creates an empty file in the proper directory, so that a file can be used
        as a reference in the Data object's output/descriptor.
        """
        root = location.get_path()
        try:
            os.makedirs(root)
            self.test_files.add(root)
        except OSError:
            pass

        filename = location.get_path(filename=filename)
        with open(filename, "w"):
            self.test_files.add(filename)

    def setUp(self):
        super().setUp()

        self.test_files = set()

        self.user = get_user_model().objects.create(username="test_user")
        processor = Process.objects.create(
            name="Test process",
            contributor=self.user,
            output_schema=[{"name": "sample", "type": "basic:file:"}],
        )
        self.data = {
            "name": "Test data",
            "contributor": self.user,
            "process": processor,
        }

    def tearDown(self):
        for filename in self.test_files:
            if os.path.isfile(filename):
                os.remove(filename)
            elif os.path.isdir(filename):
                shutil.rmtree(filename)

        super().tearDown()

    # This patch is required so that the manager is not invoked while saving Data.
    @disable_auto_calls()
    def test_remove(self):
        completed_data = Data.objects.create(**self.data)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(completed_data)
        completed_data.status = Data.STATUS_DONE
        completed_data.output = {"sample": {"file": "test-file"}}
        self.create_test_file(completed_data.location, "test-file")
        self.create_test_file(completed_data.location, "removeme")
        completed_data.save()

        pending_data = Data.objects.create(**self.data)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(pending_data)
        self.create_test_file(pending_data.location, "test-file")
        self.create_test_file(pending_data.location, "donotremoveme")

        # Check that nothing is removed if delete is False (the default).
        with patch("resolwe.flow.utils.purge.os", wraps=os) as os_mock:
            os_mock.remove = MagicMock()
            purge.purge_all()
            os_mock.remove.assert_not_called()

        completed_data.location.purged = False
        completed_data.location.save()

        # Check that only the 'removeme' file from the completed Data objects is removed
        # and files from the second (not completed) Data objects are unchanged.
        with patch("resolwe.flow.utils.purge.os", wraps=os) as os_mock:
            os_mock.remove = MagicMock()
            purge.purge_all(delete=True)
            os_mock.remove.assert_called_once_with(
                completed_data.location.get_path(filename="removeme")
            )

        completed_data.location.purged = False
        completed_data.location.save()

        # Create dummy data directories for non-existant data objects.
        self.create_test_file(DataLocation.objects.create(subpath="990"), "dummy")
        self.create_test_file(DataLocation.objects.create(subpath="991"), "dummy")

        # Check that only the 'removeme' file from the completed Data objects is removed
        # together with directories not belonging to any data objects.
        with contextlib.ExitStack() as stack:

            os_mock = stack.enter_context(
                patch("resolwe.flow.utils.purge.os", wraps=os)
            )
            shutil_mock = stack.enter_context(
                patch("resolwe.flow.utils.purge.shutil", wraps=shutil)
            )

            os_mock.remove = MagicMock()
            shutil_mock.rmtree = MagicMock()
            purge.purge_all(delete=True)
            self.assertEqual(os_mock.remove.call_count, 1)
            self.assertEqual(shutil_mock.rmtree.call_count, 2)
            os_mock.remove.assert_called_once_with(
                os.path.join(
                    settings.FLOW_EXECUTOR["DATA_DIR"],
                    str(completed_data.location.id),
                    "removeme",
                )
            )
            shutil_mock.rmtree.assert_any_call(
                os.path.join(settings.FLOW_EXECUTOR["DATA_DIR"], "990")
            )
            shutil_mock.rmtree.assert_any_call(
                os.path.join(settings.FLOW_EXECUTOR["DATA_DIR"], "991")
            )

        completed_data.location.purged = False
        completed_data.location.save()

        # Create another data object and check that if remove is called on one object,
        # only that object's data is removed.
        another_data = Data.objects.create(**self.data)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(another_data)
        another_data.status = Data.STATUS_DONE
        another_data.output = {"sample": {"file": "test-file"}}
        self.create_test_file(another_data.location, "test-file")
        self.create_test_file(another_data.location, "removeme")
        another_data.save()

        with contextlib.ExitStack() as stack:

            os_mock = stack.enter_context(
                patch("resolwe.flow.utils.purge.os", wraps=os)
            )
            shutil_mock = stack.enter_context(
                patch("resolwe.flow.utils.purge.shutil", wraps=shutil)
            )

            os_mock.remove = MagicMock()
            purge.location_purge(location_id=another_data.location.id, delete=True)
            os_mock.remove.assert_called_once_with(
                another_data.location.get_path(filename="removeme")
            )
            shutil_mock.rmtree.assert_not_called()

    # This patch is required so that the manager is not invoked while saving Data.
    @disable_auto_calls()
    def test_remove_same_location(self):
        # Create two data objects that reference the same data location and check
        # that if one object is deleted, the data is not removed upon purge.
        # It should be removed only when location is not referenced by any data object.
        same_location_data = Data.objects.create(**self.data)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(same_location_data)
        subpath = same_location_data.location.subpath

        same_location_data.output = {"sample": {"file": "test-file"}}
        same_location_data.status = Data.STATUS_DONE
        self.create_test_file(same_location_data.location, "test-file")
        same_location_data.save()

        same_location_data_2 = Data.objects.create(**self.data)
        same_location_data.location.data.add(same_location_data_2)
        same_location_data_2.output = {"sample": {"file": "test-file"}}
        same_location_data_2.status = Data.STATUS_DONE
        same_location_data_2.save()

        not_to_be_deleted = Data.objects.create(**self.data)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(not_to_be_deleted)
        not_to_be_deleted.output = {"sample": {"file": "test-file"}}
        not_to_be_deleted.status = Data.STATUS_DONE
        self.create_test_file(not_to_be_deleted.location, "test-file")
        not_to_be_deleted.save()

        self.assertEqual(Data.objects.count(), 3)

        # Delete first object.
        same_location_data.delete()

        self.assertEqual(Data.objects.count(), 2)
        with patch(
            "resolwe.flow.utils.purge.shutil.rmtree", wraps=shutil.rmtree
        ) as rmtree_mock:
            purge.purge_all(delete=True)
            rmtree_mock.assert_not_called()

        # Delete second object.
        same_location_data_2.delete()

        self.assertEqual(Data.objects.count(), 1)
        self.assertEqual(Data.objects.first().id, not_to_be_deleted.id)
        with patch(
            "resolwe.flow.utils.purge.shutil.rmtree", wraps=shutil.rmtree
        ) as rmtree_mock:
            purge.purge_all(delete=True)
            rmtree_mock.assert_called_once_with(
                os.path.join(settings.FLOW_EXECUTOR["DATA_DIR"], subpath)
            )

    # This patch is required so that the manager is not invoked while saving Data.
    @disable_auto_calls()
    def test_remove_storage(self):
        Storage.objects.create(contributor=self.user, json={})
        Storage.objects.create(contributor=self.user, json={})

        data = Data.objects.create(**self.data)
        data.storages.add(Storage.objects.create(contributor=self.user, json={}))

        purge._storage_purge_all()
        self.assertEqual(Storage.objects.count(), 3)

        purge._storage_purge_all(delete=True)
        self.assertEqual(Storage.objects.count(), 1)



# pylint: disable=missing-docstring
from django.db import models


class TestModel(models.Model):
    class Meta:
        default_permissions = ()
        permissions = (
            ("view_testmodel", "Can view model"),
            ("edit_testmodel", "Can edit model"),
        )

    name = models.CharField(max_length=30)

    field_process_type = models.CharField(max_length=100)

    number = models.IntegerField()

    date = models.DateTimeField(null=True)


class TestModelWithDependency(models.Model):

    name = models.CharField(max_length=30)

    dependencies = models.ManyToManyField("TestDependency")

    dependency = models.ForeignKey(
        "TestDependency", related_name="main_dep", null=True, on_delete=models.SET_NULL
    )


class TestDependency(models.Model):

    name = models.CharField(max_length=30)


class TestSelfDependency(models.Model):

    name = models.CharField(max_length=30)

    dependencies = models.ManyToManyField(
        "self", symmetrical=False, related_name="parents"
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-26 04:31
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import versionfield
import resolwe.flow.models.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                (
                    "settings",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                (
                    "descriptor",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                ("descriptor_dirty", models.BooleanField(default=False)),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_collection", "Can view collection"),
                    ("edit_collection", "Can edit collection"),
                    ("share_collection", "Can share collection"),
                    ("download_collection", "Can download files from collection"),
                    ("add_collection", "Can add data objects to collection"),
                    ("owner_collection", "Is owner of the collection"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Data",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("started", models.DateTimeField(blank=True, db_index=True, null=True)),
                (
                    "finished",
                    models.DateTimeField(blank=True, db_index=True, null=True),
                ),
                (
                    "checksum",
                    models.CharField(
                        db_index=True,
                        max_length=64,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_checksum",
                                message="Checksum is exactly 40 alphanumerics",
                                regex="^[0-9a-f]{64}$",
                            )
                        ],
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("UP", "Uploading"),
                            ("RE", "Resolving"),
                            ("WT", "Waiting"),
                            ("PR", "Processing"),
                            ("OK", "Done"),
                            ("ER", "Error"),
                            ("DR", "Dirty"),
                        ],
                        default="RE",
                        max_length=2,
                    ),
                ),
                ("process_pid", models.PositiveIntegerField(blank=True, null=True)),
                ("process_progress", models.PositiveSmallIntegerField(default=0)),
                ("process_rc", models.PositiveSmallIntegerField(blank=True, null=True)),
                (
                    "process_info",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=[],
                        size=None,
                    ),
                ),
                (
                    "process_warning",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=[],
                        size=None,
                    ),
                ),
                (
                    "process_error",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=[],
                        size=None,
                    ),
                ),
                ("input", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "output",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                (
                    "descriptor",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                ("descriptor_dirty", models.BooleanField(default=False)),
                ("named_by_user", models.BooleanField(default=False)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_data", "Can view data"),
                    ("edit_data", "Can edit data"),
                    ("share_data", "Can share data"),
                    ("download_data", "Can download files from data"),
                    ("owner_data", "Is owner of the data"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DescriptorSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                (
                    "schema",
                    django.contrib.postgres.fields.jsonb.JSONField(default=list),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_descriptorschema", "Can view descriptor schema"),
                    ("edit_descriptorschema", "Can edit descriptor schema"),
                    ("share_descriptorschema", "Can share descriptor schema"),
                    ("owner_descriptorschema", "Is owner of the description schema"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Entity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                (
                    "settings",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                (
                    "descriptor",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                ("descriptor_dirty", models.BooleanField(default=False)),
                ("descriptor_completed", models.BooleanField(default=False)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                ("collections", models.ManyToManyField(to="flow.Collection")),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("data", models.ManyToManyField(to="flow.Data")),
                (
                    "descriptor_schema",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="flow.DescriptorSchema",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_entity", "Can view entity"),
                    ("edit_entity", "Can edit entity"),
                    ("share_entity", "Can share entity"),
                    ("download_entity", "Can download files from entity"),
                    ("add_entity", "Can add data objects to entity"),
                    ("owner_entity", "Is owner of the entity"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="PositionInRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "position",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="flow.Entity"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Process",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "type",
                    models.CharField(
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_type",
                                message="Type may be alphanumerics separated by colon",
                                regex="^data:[a-z0-9:]+:$",
                            )
                        ],
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        default="other",
                        max_length=200,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_category",
                                message="Category may be alphanumerics separated by colon",
                                regex="^([a-z0-9]+[:\\-])*[a-z0-9]+:$",
                            )
                        ],
                    ),
                ),
                (
                    "persistence",
                    models.CharField(
                        choices=[("RAW", "Raw"), ("CAC", "Cached"), ("TMP", "Temp")],
                        default="RAW",
                        max_length=3,
                    ),
                ),
                ("description", models.TextField(default="")),
                ("data_name", models.CharField(blank=True, max_length=200, null=True)),
                (
                    "input_schema",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=list
                    ),
                ),
                (
                    "output_schema",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=list
                    ),
                ),
                (
                    "flow_collection",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("run", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "requirements",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                (
                    "scheduling_class",
                    models.CharField(
                        choices=[("IN", "Interactive"), ("BA", "Batch")],
                        default="BA",
                        max_length=2,
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_process", "Can view process"),
                    ("share_process", "Can share process"),
                    ("owner_process", "Is owner of the process"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Relation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("label", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "collection",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="flow.Collection",
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "entities",
                    models.ManyToManyField(
                        through="flow.PositionInRelation", to="flow.Entity"
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_relation", "Can view relation"),
                    ("edit_relation", "Can edit relation"),
                    ("share_relation", "Can share relation"),
                    ("owner_relation", "Is owner of the relation"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="RelationType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("ordered", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="Storage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("json", django.contrib.postgres.fields.jsonb.JSONField()),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="flow.Data"
                    ),
                ),
            ],
            options={
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="relation",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="flow.RelationType"
            ),
        ),
        migrations.AddField(
            model_name="positioninrelation",
            name="relation",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="flow.Relation"
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="descriptor_schema",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="flow.DescriptorSchema",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="parents",
            field=models.ManyToManyField(related_name="children", to="flow.Data"),
        ),
        migrations.AddField(
            model_name="data",
            name="process",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="flow.Process"
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="data",
            field=models.ManyToManyField(to="flow.Data"),
        ),
        migrations.AddField(
            model_name="collection",
            name="descriptor_schema",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="flow.DescriptorSchema",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="storage", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="relation", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="process", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="entity", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="descriptorschema", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="data", unique_together=set([("slug", "version")]),
        ),
        migrations.AlterUniqueTogether(
            name="collection", unique_together=set([("slug", "version")]),
        ),
    ]



"""SECCOMP policy.

This policy is based on the default Docker SECCOMP policy profile. It
allows several syscalls, which are most commonly used. We make one major
change regarding the network-related ``socket`` syscall in that we only
allow AF_INET/AF_INET6 SOCK_DGRAM/SOCK_STREAM sockets for TCP and UDP
protocols.
"""
# pylint: disable=too-many-lines
SECCOMP_POLICY = {
    "defaultAction": "SCMP_ACT_ERRNO",
    "architectures": ["SCMP_ARCH_X86_64", "SCMP_ARCH_X86", "SCMP_ARCH_X32"],
    "syscalls": [
        {"name": "accept", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "accept4", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "access", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "alarm", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "bind", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "brk", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "capget", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "capset", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "chdir", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "chmod", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "chown", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "chown32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "clock_getres", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "clock_gettime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "clock_nanosleep", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "close", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "connect", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "copy_file_range", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "creat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "dup", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "dup2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "dup3", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_create", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_create1", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_ctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_ctl_old", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_pwait", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_wait", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "epoll_wait_old", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "eventfd", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "eventfd2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "execve", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "execveat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "exit", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "exit_group", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "faccessat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fadvise64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fadvise64_64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fallocate", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fanotify_mark", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchdir", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchmod", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchmodat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchown", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchown32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fchownat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fcntl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fcntl64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fdatasync", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fgetxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "flistxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "flock", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fork", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fremovexattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fsetxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fstat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fstat64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fstatat64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fstatfs", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fstatfs64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "fsync", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ftruncate", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ftruncate64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "futex", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "futimesat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getcpu", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getcwd", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getdents", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getdents64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getegid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getegid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "geteuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "geteuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getgid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getgroups", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getgroups32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getitimer", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getpeername", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getpgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getpgrp", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getpid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getppid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getpriority", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getrandom", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getresgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getresgid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getresuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getresuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getrlimit", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "get_robust_list", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getrusage", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getsid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getsockname", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getsockopt", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "get_thread_area", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "gettid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "gettimeofday", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "getxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "inotify_add_watch", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "inotify_init", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "inotify_init1", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "inotify_rm_watch", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "io_cancel", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ioctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "io_destroy", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "io_getevents", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ioprio_get", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ioprio_set", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "io_setup", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "io_submit", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ipc", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "kill", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lchown", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lchown32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lgetxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "link", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "linkat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "listen", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "listxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "llistxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "_llseek", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lremovexattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lseek", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lsetxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lstat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "lstat64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "madvise", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "memfd_create", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mincore", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mkdir", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mkdirat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mknod", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mknodat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mlock", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mlock2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mlockall", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mmap", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mmap2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mprotect", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_getsetattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_notify", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_open", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_timedreceive", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_timedsend", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mq_unlink", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "mremap", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "msgctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "msgget", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "msgrcv", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "msgsnd", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "msync", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "munlock", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "munlockall", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "munmap", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "nanosleep", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "newfstatat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "_newselect", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "open", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "openat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pause", "action": "SCMP_ACT_ALLOW", "args": []},
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [{"index": 0, "value": 0, "valueTwo": 0, "op": "SCMP_CMP_EQ"}],
        },
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [{"index": 0, "value": 8, "valueTwo": 0, "op": "SCMP_CMP_EQ"}],
        },
        {
            "name": "personality",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {"index": 0, "value": 4294967295, "valueTwo": 0, "op": "SCMP_CMP_EQ"}
            ],
        },
        {"name": "pipe", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pipe2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "poll", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ppoll", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "prctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pread64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "preadv", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "prlimit64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pselect6", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pwrite64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "pwritev", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "read", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "readahead", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "readlink", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "readlinkat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "readv", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "recv", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "recvfrom", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "recvmmsg", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "recvmsg", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "remap_file_pages", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "removexattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rename", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "renameat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "renameat2", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "restart_syscall", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rmdir", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigaction", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigpending", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigprocmask", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigqueueinfo", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigreturn", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigsuspend", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_sigtimedwait", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "rt_tgsigqueueinfo", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_getaffinity", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_getattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_getparam", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_get_priority_max", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_get_priority_min", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_getscheduler", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_rr_get_interval", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_setaffinity", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_setattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_setparam", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_setscheduler", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sched_yield", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "seccomp", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "select", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "semctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "semget", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "semop", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "semtimedop", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "send", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sendfile", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sendfile64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sendmmsg", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sendmsg", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sendto", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setfsgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setfsgid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setfsuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setfsuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setgid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setgroups", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setgroups32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setitimer", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setpgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setpriority", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setregid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setregid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setresgid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setresgid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setresuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setresuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setreuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setreuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setrlimit", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "set_robust_list", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setsid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setsockopt", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "set_thread_area", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "set_tid_address", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setuid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setuid32", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "setxattr", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "shmat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "shmctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "shmdt", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "shmget", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "shutdown", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sigaltstack", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "signalfd", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "signalfd4", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sigreturn", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "socketpair", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "splice", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "stat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "stat64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "statfs", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "statfs64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "symlink", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "symlinkat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sync", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sync_file_range", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "syncfs", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "sysinfo", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "syslog", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "tee", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "tgkill", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "time", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timer_create", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timer_delete", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timerfd_create", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timerfd_gettime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timerfd_settime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timer_getoverrun", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timer_gettime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "timer_settime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "times", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "tkill", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "truncate", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "truncate64", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "ugetrlimit", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "umask", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "uname", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "unlink", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "unlinkat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "utime", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "utimensat", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "utimes", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "vfork", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "vmsplice", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "wait4", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "waitid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "waitpid", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "write", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "writev", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "arch_prctl", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "modify_ldt", "action": "SCMP_ACT_ALLOW", "args": []},
        {"name": "chroot", "action": "SCMP_ACT_ALLOW", "args": []},
        {
            "name": "clone",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 2080505856,
                    "valueTwo": 0,
                    "op": "SCMP_CMP_MASKED_EQ",
                }
            ],
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 12,  # AF_INET
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 1,
                    "value": 1,  # SOCK_STREAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 2,
                    "value": 6,  # IPPROTO_TCP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
            ],
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {"index": 0, "value": 2, "valueTwo": 0, "op": "SCMP_CMP_EQ"},  # AF_INET
                {
                    "index": 1,
                    "value": 2,  # SOCK_DGRAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 2,
                    "value": 17,  # IPPROTO_UDP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
            ],
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 10,  # AF_INET6
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 1,
                    "value": 1,  # SOCK_STREAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 2,
                    "value": 6,  # IPPROTO_TCP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
            ],
        },
        {
            "name": "socket",
            "action": "SCMP_ACT_ALLOW",
            "args": [
                {
                    "index": 0,
                    "value": 10,  # AF_INET6
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 1,
                    "value": 2,  # SOCK_DGRAM
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
                {
                    "index": 2,
                    "value": 17,  # IPPROTO_UDP
                    "valueTwo": 0,
                    "op": "SCMP_CMP_EQ",
                },
            ],
        },
    ],
}



# Generated by Django 2.2.6 on 2019-11-11 14:03

from django.db import migrations


def delete_obsolete_perms(apps, schema_editor):
    """Delete obsolete permissions from the database."""
    Permission = apps.get_model("auth", "Permission")
    Permission.objects.filter(
        codename__in=[
            "add_collection",
            "add_entity",
            "download_collection",
            "download_entity",
            "download_data",
        ]
    ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0041_remove_download_perm"),
    ]

    operations = [migrations.RunPython(delete_obsolete_perms)]



# pylint: disable=missing-docstring
import os

from resolwe.test import ProcessTestCase, tag_process, with_docker_executor


class ArchiverProcessTestCase(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self.files_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "files")
        )

    @with_docker_executor
    @tag_process("archiver")
    def test_archiver(self):
        with self.preparation_stage():
            binary = self.run_process("upload-file", {"src": "file binary"})
            image = self.run_process("upload-image-file", {"src": "file image.png"})
            tab = self.run_process("upload-tab-file", {"src": "file tab.txt"})

        archive = self.run_process(
            "archiver", {"data": [binary.id, image.id, tab.id], "fields": ["file"]}
        )

        self.assertFileExists(archive, "archive")



# pylint: disable=missing-docstring
import io

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.management import call_command
from django.test import override_settings

from guardian.shortcuts import assign_perm, remove_perm

# Import signals manually because we ignore them in App ready for tests
from resolwe.elastic import signals  # noqa: F401
from resolwe.elastic.builder import index_builder
from resolwe.test import ElasticSearchTestCase

CUSTOM_SETTINGS = {
    "INSTALLED_APPS": settings.INSTALLED_APPS + ("resolwe.elastic.tests.test_app",),
}


@override_settings(**CUSTOM_SETTINGS)
class IndexTest(ElasticSearchTestCase):
    def setUp(self):
        apps.clear_cache()
        call_command("migrate", verbosity=0, interactive=False)

        super().setUp()

    def test_mapping_multiple_times(self):
        index_builder.create_mappings()
        index_builder.create_mappings()

    def test_indexing(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument, TestSearchIndex

        # Create new object
        test_obj = TestModel.objects.create(name="Object name", number=43)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, "Object name")
        self.assertEqual(es_objects[0].num, 43)
        self.assertEqual(es_objects[0].json["key"], "value")

        # Update existing object
        test_obj.name = "Another name"
        test_obj.save()

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, "Another name")
        self.assertEqual(es_objects[0].num, 43)
        self.assertEqual(es_objects[0].json["key"], "value")

        # Create another object
        TestModel.objects.create(name="Another object", number=3)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 2)

        # Delete object
        test_obj.delete()

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Create incorrect object (User object) and try to index it
        user_model = get_user_model()
        test_incorrect = user_model.objects.create(username="user_one")
        TestSearchIndex().build(test_incorrect)

    def test_bulk_indexing(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        first_obj = TestModel.objects.create(name="First name", number=42)
        TestModel.objects.create(name="Second name", number=43)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build empty queryset
        index_builder.build(queryset=TestModel.objects.none())
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build only the subset of queryset defined in index
        index_builder.build(queryset=TestModel.objects.filter(pk=first_obj.pk))
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build only object
        index_builder.build(obj=first_obj)
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build whole queryset defined in index
        index_builder.build()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 2)

    def test_management_commands(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import (
            TestSearchDocument,
            TestAnalyzerSearchDocument,
        )

        # Prepare test data
        TestModel.objects.create(name="Object name", number=43)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge index
        call_command("elastic_purge", verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Recreate index
        call_command("elastic_index", verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge index
        call_command("elastic_purge", verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Recreate only a specific index
        call_command("elastic_index", index=["TestAnalyzerSearchIndex"], verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge only a specific index
        call_command("elastic_purge", index=["TestAnalyzerSearchIndex"], verbosity=0)

        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        call_command("elastic_index", exclude=["TestAnalyzerSearchIndex"], verbosity=0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        call_command("elastic_index", verbosity=0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        call_command("elastic_purge", exclude=["TestAnalyzerSearchIndex"], verbosity=0)

        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Recreate an invalid index
        output = io.StringIO()

        call_command(
            "elastic_index", index=["InvalidIndex"], verbosity=0, stderr=output
        )
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        call_command(
            "elastic_index", exclude=["InvalidIndex"], verbosity=0, stderr=output
        )
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        # Purge an invalid index
        call_command(
            "elastic_purge", index=["InvalidIndex"], verbosity=0, stderr=output
        )
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        call_command(
            "elastic_purge", exclude=["InvalidIndex"], verbosity=0, stderr=output
        )
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        # Create mappings.
        call_command("elastic_mapping", verbosity=0)

    def test_permissions(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        # Prepare users and groups
        user_model = get_user_model()
        user_1 = user_model.objects.create(username="user_one")
        user_2 = user_model.objects.create(username="user_two")
        user_3 = user_model.objects.create(username="user_three")
        user_pub = user_model.objects.get(username="public")
        group = Group.objects.create(name="group")

        # Create test object
        test_obj = TestModel.objects.create(name="Object name", number=43)
        assign_perm("view_testmodel", user_1, test_obj)
        assign_perm("view_testmodel", user_2, test_obj)
        assign_perm("view_testmodel", group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_2.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add user permission
        assign_perm("view_testmodel", user_3, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_2.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Remove user permission
        remove_perm("view_testmodel", user_2, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Remove group permission
        remove_perm("view_testmodel", group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add group permission
        assign_perm("view_testmodel", group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add public permission
        assign_perm("view_testmodel", AnonymousUser(), test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_pub.pk, user_1.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, True)

        # Remove public permission
        remove_perm("view_testmodel", AnonymousUser(), test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertCountEqual(
            es_objects[0].users_with_permissions, [user_1.pk, user_3.pk]
        )
        self.assertCountEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

    def test_field_name(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        TestModel.objects.create(name="Hello world FOO_BAR-G17-SA", number=42)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="hello").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="world").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search()
            .query("match", **{"field_name.raw": "hello"})
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        es_objects = (
            TestSearchDocument.search()
            .query("match", **{"field_name.raw": "Hello world FOO_BAR-G17-SA"})
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="foo").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="bar").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="g17").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="g17-sa").execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestSearchDocument.search().query("match", field_name="17").execute()
        )
        self.assertEqual(len(es_objects), 1)

    def test_field_process_type(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        TestModel.objects.create(field_process_type="data:geneset", number=42)
        TestModel.objects.create(field_process_type="data:geneset:venn", number=42)
        TestModel.objects.create(field_process_type="data:geneset:venn:omg", number=42)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 3)

        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="data")
            .execute()
        )
        self.assertEqual(len(es_objects), 3)

        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="data:geneset")
            .execute()
        )
        self.assertEqual(len(es_objects), 3)

        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="data:geneset:venn")
            .execute()
        )
        self.assertEqual(len(es_objects), 2)

        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="data:geneset:venn:omg")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        # Check if tokenizer did not include intermediate terms.
        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="geneset")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)
        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="venn")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)
        es_objects = (
            TestSearchDocument.search()
            .query("match", field_process_type="omg")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

    def test_dependencies(self):
        from .test_app.models import TestModelWithDependency, TestDependency
        from .test_app.elastic_indexes import (
            TestModelWithDependencyDocument,
            TestModelWithFilterDependencyDocument,
        )

        model = TestModelWithDependency.objects.create(name="Deps")
        dep1 = TestDependency.objects.create(name="one")
        dep2 = TestDependency.objects.create(name="two")
        dep3 = TestDependency.objects.create(name="three")
        main_dep = TestDependency.objects.create(name="main")
        model.dependency = main_dep
        model.save()
        model.dependencies.add(dep1)
        model.dependencies.add(dep2)
        dep3.testmodelwithdependency_set.add(model)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="deps")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="one")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="two")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="three")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", dependency_name="main")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        main_dep.name = "four"
        main_dep.save()

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", dependency_name="main")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", dependency_name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        main_dep.delete()

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", dependency_name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        dep3.name = "four"
        dep3.save()

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="three")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        dep3.delete()
        es_objects = (
            TestModelWithDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        # Ensure that previous updates did not cause the filtered version to be updated.
        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # If the filtered version would be updated, this would instead equal 'Deps: one, two'.
        self.assertEqual(es_objects[0].name, "Deps: ")

        dep4 = TestDependency.objects.create(name="hello")
        dep5 = TestDependency.objects.create(name="hello")
        model.dependencies.add(dep4)
        dep5.testmodelwithdependency_set.add(model)

        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # It is correct that even non-dependencies are contained in the name as dependencies are
        # only used to determine when to trigger updates.
        self.assertEqual(es_objects[0].name, "Deps: one, two, hello, hello")

        model.dependencies.remove(dep4)
        dep5.testmodelwithdependency_set.remove(model)

        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # It is correct that even non-dependencies are contained in the name as dependencies are
        # only used to determine when to trigger updates.
        self.assertEqual(es_objects[0].name, "Deps: one, two")

    def test_dependencies_reverse(self):
        from .test_app.models import TestModelWithDependency, TestDependency
        from .test_app.elastic_indexes import TestModelWithReverseDependencyDocument

        dep = TestDependency.objects.create(name="deps")
        model1 = TestModelWithDependency.objects.create(name="One", dependency=dep)
        model2 = TestModelWithDependency.objects.create(name="Two")
        model3 = TestModelWithDependency.objects.create(name="Three")
        model1.dependencies.add(dep)
        model2.dependencies.add(dep)
        dep.main_dep.add(model2)
        dep.testmodelwithdependency_set.add(model3)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="deps")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="one")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="two")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="three")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        model3.name = "Four"
        model3.save()

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        model3.delete()

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", main_dep_name="two")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", main_dep_name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

        model2.name = "four"
        model2.save()

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", main_dep_name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)

        model2.delete()

        es_objects = (
            TestModelWithReverseDependencyDocument.search()
            .query("match", main_dep_name="four")
            .execute()
        )
        self.assertEqual(len(es_objects), 0)

    def test_dependencies_self(self):
        from .test_app.models import TestSelfDependency
        from .test_app.elastic_indexes import TestModelWithSelfDependencyDocument

        dep1 = TestSelfDependency.objects.create(name="One")
        dep2 = TestSelfDependency.objects.create(name="Two")
        dep3 = TestSelfDependency.objects.create(name="Three")
        parent = TestSelfDependency.objects.create(name="Parent")
        parent.dependencies.add(dep1)
        parent.dependencies.add(dep2)
        dep3.parents.add(parent)

        es_objects = (
            TestModelWithSelfDependencyDocument.search()
            .query("match", name="parent")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, "Parent: One, Two, Three")

        dep2.name = "Too"
        dep2.save()

        es_objects = (
            TestModelWithSelfDependencyDocument.search()
            .query("match", name="parent")
            .execute()
        )
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, "Parent: One, Too, Three")



""".. Ignore pydocstyle D400.

=====================
Elastic Index Builder
=====================

"""
import inspect
import os
import re
import threading
import uuid
from importlib import import_module

from django.apps import apps
from django.db import models
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
)
from django.db.models.signals import m2m_changed, post_delete, post_save, pre_delete

from resolwe.test.utils import is_testing

from .composer import composer
from .indices import BaseIndex
from .utils import prepare_connection

__all__ = (
    "index_builder",
    "ManyToManyDependency",
)

# UUID used in tests to make sure that no index is re-used.
TESTING_UUID = str(uuid.uuid4())


class BuildArgumentsCache:
    """Cache for storing arguments for index builder.

    If set value contains a queryset, it is evaluated and later
    recreated to prevent influence from database changes.

    """

    _thread_local = threading.local()

    def _get_cache_key(self, obj):
        """Derive cache key for given object."""
        if obj is not None:
            # Make sure that key is REALLY unique.
            return "{}-{}".format(id(self), obj.pk)

        return "{}-None".format(id(self))

    def _clean_cache(self, obj):
        """Clean cache."""
        del self._thread_local.cache[self._get_cache_key(obj)]

    def set(self, obj, build_kwargs):
        """Set cached value."""
        if build_kwargs is None:
            build_kwargs = {}

        cached = {}
        if "queryset" in build_kwargs:
            cached = {
                "model": build_kwargs["queryset"].model,
                "pks": list(build_kwargs["queryset"].values_list("pk", flat=True)),
            }

        elif "obj" in build_kwargs:
            cached = {
                "obj": build_kwargs["obj"],
            }

        if not hasattr(self._thread_local, "cache"):
            self._thread_local.cache = {}
        self._thread_local.cache[self._get_cache_key(obj)] = cached

    def take(self, obj):
        """Get cached value and clean cache."""
        cached = self._thread_local.cache[self._get_cache_key(obj)]
        build_kwargs = {}

        if "model" in cached and "pks" in cached:
            build_kwargs["queryset"] = cached["model"].objects.filter(
                pk__in=cached["pks"]
            )

        elif "obj" in cached:
            if cached["obj"].__class__.objects.filter(pk=cached["obj"].pk).exists():
                build_kwargs["obj"] = cached["obj"]
            else:
                # Object was deleted in the meantime.
                build_kwargs["queryset"] = cached["obj"].__class__.objects.none()

        self._clean_cache(obj)

        return build_kwargs


class ElasticSignal:
    """Class for creating signals to update/create indexes.

    To register the signal, add the following code::

        signal = ElasticSignal(<my_signal>, <method_name>)
        signal.connect(<signal_type>, [sender=<my_model>])

    You may later disconnect the signal by calling::

        signal.disconnect()

    ``signal type`` can be i.e. ``django.db.models.signals.pre_save``.

    """

    def __init__(self, index, method_name, pass_kwargs=False):
        """Initialize signal."""
        self.index = index
        self.method_name = method_name
        self.pass_kwargs = pass_kwargs
        self.connections = []

    def connect(self, signal, **kwargs):
        """Connect a specific signal type to this receiver."""
        signal.connect(self, **kwargs)
        self.connections.append((signal, kwargs))

    def disconnect(self):
        """Disconnect all connected signal types from this receiver."""
        for signal, kwargs in self.connections:
            signal.disconnect(self, **kwargs)

    def __call__(self, sender, instance, **kwargs):
        """Process signal."""
        method = getattr(self.index, self.method_name)
        if self.pass_kwargs:
            method(obj=instance, **kwargs)
        else:
            method(obj=instance)


class Dependency:
    """Abstract base class for index model dependencies."""

    def __init__(self, model):
        """Construct dependency."""
        self.model = model
        self.index = None

        # Cache for pre/post-delete signals.
        self.delete_cache = BuildArgumentsCache()

    def connect(self, index):
        """Connect signals needed for dependency updates.

        Pre- and post-delete signals have to be handled separately, as:

          * in the pre-delete signal we have the information which
            objects to rebuild, but affected relations are still
            presented, so rebuild would reflect in the wrong (outdated)
            indices
          * in the post-delete signal indices can be rebuild corectly,
            but there is no information which objects to rebuild, as
            affected relations were already deleted

        To bypass this, list of objects should be stored in the
        pre-delete signal indexing should be triggered in the
        post-delete signal.
        """
        self.index = index

        signal = ElasticSignal(self, "process", pass_kwargs=True)
        signal.connect(post_save, sender=self.model)
        signal.connect(pre_delete, sender=self.model)

        pre_delete_signal = ElasticSignal(self, "process_predelete", pass_kwargs=True)
        pre_delete_signal.connect(pre_delete, sender=self.model)

        post_delete_signal = ElasticSignal(self, "process_delete", pass_kwargs=True)
        post_delete_signal.connect(post_delete, sender=self.model)

        return [signal, pre_delete_signal, post_delete_signal]

    def filter(self, obj, update_fields=None):
        """Determine if dependent object should be processed.

        If ``False`` is returned, processing of the dependent object will
        be aborted.
        """

    def _filter(self, objects, **kwargs):
        """Determine if dependent object should be processed."""
        for obj in objects:
            if self.filter(obj, **kwargs) is False:
                return False

        return True

    def _get_build_kwargs(self, **kwargs):
        """Prepare arguments for rebuilding indices."""
        raise NotImplementedError

    def process_predelete(self, obj, **kwargs):
        """Render the queryset of influenced objects and cache it."""
        build_kwargs = self._get_build_kwargs(obj, **kwargs)
        self.delete_cache.set(obj, build_kwargs)

    def process_delete(self, obj, **kwargs):
        """Recreate queryset from the index and rebuild the index."""
        build_kwargs = self.delete_cache.take(obj)

        if build_kwargs:
            self.index.build(**build_kwargs)

    def process(self, obj, **kwargs):
        """Process signals from dependencies."""
        build_kwargs = self._get_build_kwargs(obj, **kwargs)

        if build_kwargs:
            self.index.build(**build_kwargs)


class ManyToManyDependency(Dependency):
    """Dependency on a many-to-many relation."""

    def __init__(self, field):
        """Construct m2m dependency."""
        # We use None as the model as we cannot determine it until assigned to an index.
        super().__init__(None)

        self.accessor = None
        self.field = field
        # Cache for pre/post-remove action in m2m_changed signal.
        self.remove_cache = BuildArgumentsCache()

    def connect(self, index):
        """Connect signals needed for dependency updates."""
        # Determine which model is the target model as either side of the relation
        # may be passed as `field`.
        if index.object_type == self.field.rel.model:
            self.model = self.field.rel.related_model
            self.accessor = self.field.rel.field.attname
        else:
            self.model = self.field.rel.model
            if self.field.rel.symmetrical:
                # Symmetrical m2m relation on self has no reverse accessor.
                raise NotImplementedError(
                    "Dependencies on symmetrical M2M relations are not supported due "
                    "to strange handling of the m2m_changed signal which only makes "
                    "half of the relation visible during signal execution. For now you "
                    "need to use symmetrical=False on the M2M field definition."
                )
            else:
                self.accessor = self.field.rel.get_accessor_name()

        # Connect signals.
        signals = super().connect(index)

        m2m_signal = ElasticSignal(self, "process_m2m", pass_kwargs=True)
        m2m_signal.connect(m2m_changed, sender=self.field.through)
        signals.append(m2m_signal)

        # If the relation has a custom through model, we need to subscribe to it.
        if not self.field.rel.through._meta.auto_created:
            signal = ElasticSignal(self, "process_m2m_through_save", pass_kwargs=True)
            signal.connect(post_save, sender=self.field.rel.through)
            signals.append(signal)

            signal = ElasticSignal(
                self, "process_m2m_through_pre_delete", pass_kwargs=True
            )
            signal.connect(pre_delete, sender=self.field.rel.through)
            signals.append(signal)

            signal = ElasticSignal(
                self, "process_m2m_through_post_delete", pass_kwargs=True
            )
            signal.connect(post_delete, sender=self.field.rel.through)
            signals.append(signal)

        return signals

    def _get_build_kwargs(
        self, obj, pk_set=None, action=None, update_fields=None, reverse=None, **kwargs
    ):
        """Prepare arguments for rebuilding indices."""
        if action is None:
            # Check filter before rebuilding index.
            if not self._filter([obj], update_fields=update_fields):
                return

            queryset = getattr(obj, self.accessor).all()

            # Special handling for relations to self.
            if self.field.rel.model == self.field.rel.related_model:
                queryset = queryset.union(
                    getattr(obj, self.field.rel.get_accessor_name()).all()
                )

            return {"queryset": queryset}
        else:
            # Update to relation itself, only update the object in question.
            if self.field.rel.model == self.field.rel.related_model:
                # Special case, self-reference, update both ends of the relation.
                pks = set()
                if self._filter(self.model.objects.filter(pk__in=pk_set)):
                    pks.add(obj.pk)

                if self._filter(self.model.objects.filter(pk__in=[obj.pk])):
                    pks.update(pk_set)

                return {"queryset": self.index.object_type.objects.filter(pk__in=pks)}
            elif isinstance(obj, self.model):
                # Need to switch the role of object and pk_set.
                result = {
                    "queryset": self.index.object_type.objects.filter(pk__in=pk_set)
                }
                pk_set = {obj.pk}
            else:
                result = {"obj": obj}

            if action != "post_clear":
                # Check filter before rebuilding index.
                if not self._filter(self.model.objects.filter(pk__in=pk_set)):
                    return

            return result

    def process_m2m(
        self,
        obj,
        pk_set=None,
        action=None,
        update_fields=None,
        cache_key=None,
        **kwargs,
    ):
        """Process signals from dependencies.

        Remove signal is processed in two parts. For details see:
        :func:`~Dependency.connect`
        """
        if action not in (None, "post_add", "pre_remove", "post_remove", "post_clear"):
            return

        if action == "post_remove":
            build_kwargs = self.remove_cache.take(cache_key)
        else:
            build_kwargs = self._get_build_kwargs(
                obj, pk_set, action, update_fields, **kwargs
            )

        if action == "pre_remove":
            self.remove_cache.set(cache_key, build_kwargs)
            return

        if build_kwargs:
            self.index.build(**build_kwargs)

    def _process_m2m_through(self, obj, action):
        """Process custom M2M through model actions."""
        source = getattr(obj, self.field.rel.field.m2m_field_name())
        target = getattr(obj, self.field.rel.field.m2m_reverse_field_name())

        pk_set = set()
        if target:
            pk_set.add(target.pk)

        self.process_m2m(source, pk_set, action=action, reverse=False, cache_key=obj)

    def process_m2m_through_save(self, obj, created=False, **kwargs):
        """Process M2M post save for custom through model."""
        # We are only interested in signals that establish relations.
        if not created:
            return

        self._process_m2m_through(obj, "post_add")

    def process_m2m_through_pre_delete(self, obj, **kwargs):
        """Process M2M pre delete for custom through model."""
        self._process_m2m_through(obj, "pre_remove")

    def process_m2m_through_post_delete(self, obj, **kwargs):
        """Process M2M post delete for custom through model."""
        self._process_m2m_through(obj, "post_remove")


class ReverseManyToOneDependency(Dependency):
    """Dependency on a reverse many-to-one relation."""

    def __init__(self, descriptor):
        """Construct reverse m2o dependency."""
        super().__init__(descriptor.rel.related_model)
        self.accessor = descriptor.rel.field.attname

    def _get_build_kwargs(self, obj, update_fields=None, **kwargs):
        """Prepare arguments for rebuilding indices."""
        if not self._filter([obj], update_fields=update_fields):
            return

        return {
            "queryset": self.index.object_type.objects.filter(
                pk=getattr(obj, self.accessor)
            )
        }


class ForwardManyToOneDependency(Dependency):
    """Dependency on a forward many-to-one relation."""

    def __init__(self, descriptor):
        """Construct forward m2o dependency."""
        super().__init__(descriptor.field.related_model)
        self.accessor = descriptor.field.remote_field.get_accessor_name()

    def _get_build_kwargs(self, obj, update_fields=None, **kwargs):
        """Prepare arguments for rebuilding indices."""
        if not self._filter([obj], update_fields=update_fields):
            return

        return {"queryset": getattr(obj, self.accessor).all()}


class IndexBuilder:
    """Find indexes and register corresponding signals.

    Indexes are auto collected from ``elastic_indexes.py`` files from
    all Django registered apps

    Post-save and pre-delete signals are registered for objects
    specified in ``object_type`` attribute of each index.

    """

    def __init__(self):
        """Initialize index builder object."""
        #: list of index builders
        self.indexes = []

        #: list of registered signals
        self.signals = []

        prepare_connection()

        self.discover_indexes()
        self.register_signals()

    def _connect_signal(self, index):
        """Create signals for building indexes."""
        post_save_signal = ElasticSignal(index, "build")
        post_save_signal.connect(post_save, sender=index.object_type)
        self.signals.append(post_save_signal)

        post_delete_signal = ElasticSignal(index, "remove_object")
        post_delete_signal.connect(post_delete, sender=index.object_type)
        self.signals.append(post_delete_signal)

        # Connect signals for all dependencies.
        for dependency in index.get_dependencies():
            # Automatically convert m2m fields to dependencies.
            if isinstance(dependency, (models.ManyToManyField, ManyToManyDescriptor)):
                dependency = ManyToManyDependency(dependency)
            elif isinstance(dependency, ReverseManyToOneDescriptor):
                dependency = ReverseManyToOneDependency(dependency)
            elif isinstance(dependency, ForwardManyToOneDescriptor):
                dependency = ForwardManyToOneDependency(dependency)
            elif not isinstance(dependency, Dependency):
                raise TypeError(
                    "Unsupported dependency type: {}".format(repr(dependency))
                )

            signal = dependency.connect(index)
            self.signals.extend(signal)

    def unregister_signals(self):
        """Delete signals for building indexes."""
        for signal in self.signals:
            signal.disconnect()
        self.signals = []

    def register_signals(self):
        """Register signals for all indexes."""
        for index in self.indexes:
            if index.object_type:
                self._connect_signal(index)

    def create_mappings(self):
        """Create mappings for all indexes."""
        for index in self.indexes:
            index.create_mapping()

    def discover_indexes(self):
        """Save list of index builders into ``_index_builders``."""
        self.indexes = []

        for app_config in apps.get_app_configs():
            indexes_path = "{}.elastic_indexes".format(app_config.name)
            try:
                indexes_module = import_module(indexes_path)

                for attr_name in dir(indexes_module):
                    attr = getattr(indexes_module, attr_name)
                    if (
                        inspect.isclass(attr)
                        and issubclass(attr, BaseIndex)
                        and attr is not BaseIndex
                    ):
                        # Make sure that parallel tests have different indices.
                        if is_testing():
                            index = attr.document_class._index._name
                            testing_postfix = "_test_{}_{}".format(
                                TESTING_UUID, os.getpid()
                            )

                            if not index.endswith(testing_postfix):
                                # Replace current postfix with the new one.
                                if attr.testing_postfix:
                                    index = index[: -len(attr.testing_postfix)]
                                index = index + testing_postfix
                                attr.testing_postfix = testing_postfix

                            attr.document_class._index._name = index

                        index = attr()

                        # Apply any extensions defined for the given index. Currently index extensions are
                        # limited to extending "mappings".
                        for extension in composer.get_extensions(attr):
                            mapping = getattr(extension, "mapping", {})
                            index.mapping.update(mapping)

                        self.indexes.append(index)
            except ImportError as ex:
                if not re.match("No module named .*elastic_indexes.*", str(ex)):
                    raise

    def build(self, obj=None, queryset=None, push=True):
        """Trigger building of the indexes.

        Support passing ``obj`` parameter to the indexes, so we can
        trigger build only for one object.
        """
        for index in self.indexes:
            index.build(obj, queryset, push)

    def push(self, index=None):
        """Push built documents to ElasticSearch.

        If ``index`` is specified, only that index will be pushed.
        """
        for ind in self.indexes:
            if index and not isinstance(ind, index):
                continue
            ind.push()

    def delete(self, skip_mapping=False):
        """Delete all entries from ElasticSearch."""
        for index in self.indexes:
            index.destroy()
            if not skip_mapping:
                index.create_mapping()

    def remove_object(self, obj):
        """Delete given object from all indexes."""
        for index in self.indexes:
            index.remove_object(obj)

    def destroy(self):
        """Delete all indexes from Elasticsearch and index builder."""
        self.unregister_signals()
        for index in self.indexes:
            index.destroy()
        self.indexes = []


index_builder = IndexBuilder()



"""Elastic Search indexes for Data model."""
import elasticsearch_dsl as dsl

from resolwe.elastic.fields import Name, ProcessType, Slug
from resolwe.elastic.indices import BaseIndex

from ..models import Data
from .base import BaseDocument, BaseIndexMixin


class DataDocument(BaseDocument):
    """Document for data search."""

    started = dsl.Date()
    finished = dsl.Date()
    status = dsl.Keyword()
    process = dsl.Integer()
    process_type = ProcessType()
    # Keep backward compatibility.
    type = ProcessType()
    process_name = Name()
    process_slug = Slug()
    tags = dsl.Keyword(multi=True)

    collection = dsl.Integer()
    entity = dsl.Integer()

    class Index:
        """Meta class for data search document."""

        name = "data"


class DataIndex(BaseIndexMixin, BaseIndex):
    """Index for data objects used in ``DataDocument``."""

    queryset = Data.objects.all().prefetch_related("process", "contributor")
    object_type = Data
    document_class = DataDocument

    mapping = {
        "process": "process.id",
        "process_name": "process.name",
        "process_type": "process.type",
        "process_slug": "process.slug",
        "type": "process.type",
        "entity": "entity_id",
        "collection": "collection_id",
    }



""".. Ignore pydocstyle D400.

===============
Elastic Indices
===============
"""
from resolwe.elastic.builder import index_builder


class ElasticIndexFilterMixin:
    """Mixin for management commands with index filtering."""

    def add_arguments(self, parser):
        """Command arguments."""
        super().add_arguments(parser)

        parser.add_argument(
            "-i",
            "--index",
            action="append",
            default=[],
            help="Only process specific indices (default: not set)",
        )
        parser.add_argument(
            "-e",
            "--exclude",
            action="append",
            default=[],
            help="Do not process specific indices (default: not set)",
        )

    def invalid_index(self, name):
        """Show an invalid index error message."""
        self.stderr.write("Unknown index: {}".format(name))
        self.stderr.write("Supported indices are:")
        for index in index_builder.indexes:
            self.stderr.write("  * {}".format(index.__class__.__name__))

    def has_filter(self, options):
        """Return true if any filters have been specified."""
        return options["index"] or options["exclude"]

    def handle_index(self, index):
        """Process index."""

    def filter_indices(self, options, verbosity, *args, **kwargs):
        """Filter indices and execute an action for each index."""
        index_name_map = {
            index.__class__.__name__: index for index in index_builder.indexes
        }

        # Process includes.
        if options["index"]:
            indices = set(options["index"])
        else:
            indices = set(index_name_map.keys())

        # Process excludes.
        for index_name in options["exclude"]:
            if index_name not in index_name_map:
                self.invalid_index(index_name)
                return

            indices.discard(index_name)

        # Execute action for each remaining index.
        for index_name in indices:
            try:
                index = index_name_map[index_name]
            except KeyError:
                self.invalid_index(index_name)
                return

            if verbosity > 0:
                self.stdout.write("Processing index '{}'...".format(index_name))
            self.handle_index(index, *args, **kwargs)



""".. Ignore pydocstyle D400.

======================
Command: elastic_purge
======================

"""
from django.core.management.base import BaseCommand

from resolwe.elastic.builder import index_builder
from resolwe.elastic.mixins import ElasticIndexFilterMixin


class Command(ElasticIndexFilterMixin, BaseCommand):
    """Purge ElasticSearch indexes."""

    help = "Purge ElasticSearch indexes."

    def add_arguments(self, parser):
        """Add an argument to specify output format."""
        super().add_arguments(parser)

        parser.add_argument(
            "--skip-mapping",
            dest="skip_mapping",
            action="store_true",
            help="Don't push fresh mappings to Elasticserch after deleting indices",
        )

    def handle_index(self, index, skip_mapping):
        """Process index."""
        index.destroy()
        if not skip_mapping:
            index.create_mapping()

    def handle(self, *args, **options):
        """Command handle."""
        verbosity = int(options["verbosity"])
        skip_mapping = options["skip_mapping"]

        if self.has_filter(options):
            self.filter_indices(options, verbosity, skip_mapping=skip_mapping)
        else:
            # Process all indices.
            index_builder.delete(skip_mapping=skip_mapping)



"""Protocol constants used by the manager and executors."""


class WorkerProtocol:
    """Constants used by the manager workers."""

    COMMAND = "command"
    DATA_ID = "data_id"

    COMMUNICATE = "communicate"
    COMMUNICATE_SETTINGS = "settings_override"
    COMMUNICATE_EXTRA = "kwargs"

    FINISH = "finish_data"
    FINISH_SPAWNED = "spawned"
    FINISH_COMMUNICATE_EXTRA = "communicate_kwargs"

    ABORT = "abort_data"


class ExecutorProtocol:
    """Constants used by the executor<->listener protocol."""

    COMMAND = "command"
    DATA_ID = "data_id"

    UPDATE = "update"
    UPDATE_CHANGESET = "changeset"

    FINISH = "finish"
    FINISH_PROCESS_RC = "process_rc"
    FINISH_SPAWN_PROCESSES = "spawn_processes"
    FINISH_EXPORTED_FILES = "exported_files_mapper"

    ABORT = "abort"

    RESULT = "result"
    RESULT_OK = "OK"
    RESULT_ERROR = "ER"

    LOG = "log"
    LOG_MESSAGE = "message"


class ExecutorFiles:
    """Various files used by the executor."""

    FILE_LIST_KEY = "serialized_files"

    EXECUTOR_SETTINGS = "settings.json"
    DJANGO_SETTINGS = "django_settings.json"
    DATA = "data.json"
    DATA_LOCATION = "data_location.json"
    DATA_META = "data_meta.json"
    PROCESS = "process.json"
    PROCESS_META = "process_meta.json"

    PROCESS_SCRIPT = "process_script.sh"

    SECRETS_DIR = "secrets"



# pylint: disable=missing-docstring
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse

from guardian.shortcuts import assign_perm
from rest_framework.test import APITestCase

from resolwe.flow.models import Process


class ProcessOrderingTest(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        user = user_model.objects.create(username="user")

        self.proc_1 = Process.objects.create(
            name="My process", contributor=user, version=1
        )
        self.proc_2 = Process.objects.create(
            name="My process", contributor=user, version=2
        )

        assign_perm("view_process", AnonymousUser(), self.proc_1)
        assign_perm("view_process", AnonymousUser(), self.proc_2)

        self.url = reverse("resolwe-api:process-list")

    def test_ordering_version(self):
        response = self.client.get(self.url, {"ordering": "version"}, format="json")
        self.assertEqual(response.data[0]["id"], self.proc_1.id)
        self.assertEqual(response.data[1]["id"], self.proc_2.id)

        response = self.client.get(self.url, {"ordering": "-version"}, format="json")
        self.assertEqual(response.data[0]["id"], self.proc_2.id)
        self.assertEqual(response.data[1]["id"], self.proc_1.id)



"""An execution engine that outputs bash programs."""
import copy

import shellescape
import yaml

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.flow.execution_engines.exceptions import ExecutionError
from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models.utils import hydrate_input_references, hydrate_input_uploads


class SafeString(str):
    """String wrapper for marking strings safe."""


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that outputs bash programs."""

    name = "bash"

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        if not path.lower().endswith((".yml", ".yaml")):
            return []

        with open(path) as fn:
            schemas = yaml.load(fn, Loader=yaml.FullLoader)
        if not schemas:
            # TODO: Logger.
            # self.stderr.write("Could not read YAML file {}".format(schema_file))
            return []

        process_schemas = []
        for schema in schemas:
            if "run" not in schema:
                continue

            # NOTE: This currently assumes that 'bash' is the default.
            if schema["run"].get("language", "bash") != "bash":
                continue

            process_schemas.append(schema)

        return process_schemas

    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        try:
            inputs = copy.deepcopy(data.input)
            hydrate_input_references(inputs, data.process.input_schema)
            hydrate_input_uploads(inputs, data.process.input_schema)

            # Include special 'proc' variable in the context.
            inputs["proc"] = {
                "data_id": data.id,
                "data_dir": self.manager.get_executor().resolve_data_path(),
            }

            # Include special 'requirements' variable in the context.
            inputs["requirements"] = data.process.requirements
            # Inject default values and change resources according to
            # the current Django configuration.
            inputs["requirements"]["resources"] = data.process.get_resource_limits()

            script_template = data.process.run.get("program", "")

            # Get the appropriate expression engine. If none is defined, do not evaluate
            # any expressions.
            expression_engine = data.process.requirements.get("expression-engine", None)
            if not expression_engine:
                return script_template

            return self.get_expression_engine(expression_engine).evaluate_block(
                script_template, inputs, escape=self._escape, safe_wrapper=SafeString,
            )
        except EvaluationError as error:
            raise ExecutionError("{}".format(error))

    def _escape(self, value):
        """Escape given value unless it is safe."""
        if isinstance(value, SafeString):
            return value

        return shellescape.quote(value)



# pylint: disable=missing-docstring
import elasticsearch_dsl as dsl

from resolwe.elastic.builder import ManyToManyDependency
from resolwe.elastic.fields import Name, ProcessType
from resolwe.elastic.indices import BaseDocument, BaseIndex

from .models import (
    TestDependency,
    TestModel,
    TestModelWithDependency,
    TestSelfDependency,
)


class TestSearchDocument(BaseDocument):
    id = dsl.Integer()
    name = dsl.Text(fielddata=True)
    num = dsl.Integer()
    date = dsl.Date()
    json = dsl.Object()

    field_name = Name()
    field_process_type = ProcessType()
    none_test = dsl.Integer()

    class Index:
        name = "test_search"


class TestSearchIndex(BaseIndex):
    mapping = {
        "num": "number",
        "field_name": "name",
        "none_test": "this.does.not.exist",
    }

    queryset = TestModel.objects.all()
    object_type = TestModel
    document_class = TestSearchDocument

    def get_json_value(self, obj):
        return {"key": "value"}


class TestAnalyzerSearchDocument(BaseDocument):
    name = dsl.Text(
        analyzer=dsl.analyzer(
            "test_analyzer", tokenizer="keyword", filter=["lowercase",], fielddata=True,
        )
    )

    class Index:
        name = "test_analyzer_search"


class TestAnalyzerSearchIndex(BaseIndex):
    queryset = TestModel.objects.all()
    object_type = TestModel
    document_class = TestAnalyzerSearchDocument


class TestModelWithDependencyDocument(BaseDocument):
    name = dsl.Text(fielddata=True)
    dependency_name = dsl.Text(fielddata=True)

    class Index:
        name = "test_model_with_dependency_search"


class TestModelWithDependencySearchIndex(BaseIndex):
    queryset = TestModelWithDependency.objects.all().prefetch_related("dependencies")
    object_type = TestModelWithDependency
    document_class = TestModelWithDependencyDocument

    def get_dependencies(self):
        return [
            TestModelWithDependency.dependencies,
            TestModelWithDependency.dependency,
        ]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all()]
        return "{}: {}".format(obj.name, ", ".join(names))

    def get_dependency_name_value(self, obj):
        return obj.dependency.name


class TestModelWithFilterDependencyDocument(BaseDocument):
    name = dsl.Text(fielddata=True)

    class Index:
        name = "test_model_with_filter_dependency_search"


class FilterHelloDependency(ManyToManyDependency):
    def filter(self, obj, update_fields=None):
        return obj.name == "hello"


class TestModelWithFilterDependencySearchIndex(BaseIndex):
    queryset = TestModelWithDependency.objects.all().prefetch_related("dependencies")
    object_type = TestModelWithDependency
    document_class = TestModelWithFilterDependencyDocument

    def get_dependencies(self):
        return [FilterHelloDependency(TestModelWithDependency.dependencies)]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all()]
        return "{}: {}".format(obj.name, ", ".join(names))


class TestModelWithReverseDependencyDocument(BaseDocument):
    name = dsl.Text()
    main_dep_name = dsl.Text()

    class Index:
        name = "test_model_with_reverse_dependency_search"


class TestModelWithReverseDependencySearchIndex(BaseIndex):
    queryset = TestDependency.objects.all()
    object_type = TestDependency
    document_class = TestModelWithReverseDependencyDocument

    def get_dependencies(self):
        return [
            TestDependency.testmodelwithdependency_set,
            TestDependency.main_dep,
        ]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.testmodelwithdependency_set.all()]
        return "{}: {}".format(obj.name, ", ".join(names))

    def get_main_dep_name_value(self, obj):
        names = [dep.name for dep in obj.main_dep.all()]
        return "{}: {}".format(obj.name, ", ".join(names))


class TestModelWithSelfDependencyDocument(BaseDocument):
    name = dsl.Text()

    class Index:
        name = "test_model_with_self_dependency_search"


class TestModelWithSelfDependencySearchIndex(BaseIndex):
    queryset = TestSelfDependency.objects.all()
    object_type = TestSelfDependency
    document_class = TestModelWithSelfDependencyDocument

    def get_dependencies(self):
        return [TestSelfDependency.dependencies]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all().order_by("pk")]
        return "{}: {}".format(obj.name, ", ".join(names))
